import os
from importlib import import_module

from bx24_orm.core.settings import Bx24Settings

BX_SETTINGS_MODULE = 'BX24_SETTINGS_MODULE'
try:
    __settings_module = import_module(os.environ[BX_SETTINGS_MODULE])
except KeyError:
    raise RuntimeError('BX24_SETTINGS_MODULE is not specified')

__module_dict_vars = {}
for attr in dir(__settings_module):
    if not attr.startswith('__'):
        __module_dict_vars.update({attr: getattr(__settings_module, attr)})

settings = Bx24Settings(__module_dict_vars)
token_storage = settings.TOKEN_STORAGE_CLS(settings)



# -*- coding: utf-8 -*-

import requests
import six
import argparse


def __get_tokens(client_id, client_secret, domain):
    code_url = 'https://{}.bitrix24.ru/oauth/authorize/?response_type=code&client_id={}'.format(domain, client_id)
    print(code_url)
    input_func = input
    if six.PY2:
        input_func = raw_input
    code = input_func('Follow link and copy code from result url. Then enter code here: ')
    auth_url = 'https://{}.bitrix24.ru/oauth/token/'.format(domain)
    auth_parameters = {
        'client_id': client_id,
        'client_secret': client_secret,
        'code': code,
        'grant_type': 'authorization_code',
        'scope': ''
    }
    response = requests.get(auth_url, auth_parameters)
    if response.status_code == 200:
        result = response.json()
        auth_data = {
            'access_token': result['access_token'],
            'refresh_token': result['refresh_token']
        }
        print(auth_data)
        return 0
    else:
        print('Response status code {}. Try again!'.format(response.status_code))
        return -1


def bx24_cmd():
    parser = argparse.ArgumentParser()
    parser.add_argument('command', help='[get_tokens]', action='store')
    parser.add_argument('-c', dest='client_id', help='client_id', action='store', type=str, required=True)
    parser.add_argument('-s', dest='client_secret', help='client_secret', action='store', type=str, required=True)
    parser.add_argument('-d', dest='domain', help='domain', action='store', type=str, required=True)
    parsed = parser.parse_args()
    if parsed.command == 'get_tokens':
        return __get_tokens(parsed.client_id, parsed.client_secret, parsed.domain)


if __name__ == '__main__':
    bx24_cmd()



# -*- coding: utf-8 -*-

from .token_storage import DefaultTokenStorage

BX_SETTINGS_MODULE = 'BX24_SETTINGS_MODULE'


class Bx24Settings(object):
    def __init__(self, settings_dict):
        self.TOKEN_STORAGE_FILENAME = 'bx24_tokens'
        self.TOKEN_STORAGE_CLS = DefaultTokenStorage
        for k in settings_dict:
            setattr(self, k, settings_dict[k])

    @property
    def default_domain(self):
        settings_default_domain = getattr(self, 'DEFAULT_DOMAIN', None)
        if not settings_default_domain:
            raise RuntimeError('Default domain not specified')
        return settings_default_domain

    @property
    def default_transport(self):
        return getattr(self, 'DEFAULT_TRANSPORT', 'json')



# -*- coding: utf-8 -*-


class BaseBxAdapter(object):
    def __init__(self, entity_cls):
        self.entity_cls = entity_cls

    def from_bitrix(self, response):
        # type: (BxQueryResponse) -> BxEntity or list
        if response.result:
            if type(response.result) is list:
                return [self.entity_cls(**r) for r in response.result]
            else:
                return self.entity_cls(**response.result)
        else:
            return None

    @classmethod
    def to_bitrix(cls, bx_entity, force_all_fields=False):
        # type: (BxEntity, bool) -> dict
        return bx_entity.get_updates(force_all_fields)



NOT_JSON_RESPONSE = 'NOT_JSON_RESPONSE'
UNEXPECTED_HTTP_CODE = 'UNEXPECTED_HTTP_CODE'
QUERY_LIMIT_EXCEEDED = 'QUERY_LIMIT_EXCEEDED'
NO_AUTH_FOUND = 'NO_AUTH_FOUND'
ERROR_METHOD_NOT_FOUND = 'ERROR_METHOD_NOT_FOUND'
NETWORK_ERROR = 'NETWORK_ERROR'


class BxException(Exception):
    def __init__(self, error, error_description):
        super(Exception, self).__init__()
        self.error = error
        self.error_description = error_description

    def __str__(self):
        return str((self.error, self.error_description))


class NotFoundException(BxException):
    pass


class InvalidQueryException(BxException):
    pass


class InvalidTokenException(BxException):
    pass


class InvalidRequestException(BxException):
    pass


class RequestFailedException(BxException):
    pass


class QueryLimitExceededException(RequestFailedException):
    pass


class NotJsonResponseException(RequestFailedException):
    pass


class TokenExpiredException(InvalidTokenException):
    pass


class TokenNotFoundException(InvalidTokenException):
    pass


class UnexpectedResponseCodeException(RequestFailedException):
    pass



# -*- coding: utf-8 -*-
from bx24_orm.core import settings, token_storage
from bx24_orm.core.bx_interface import (BxQuery, BxBatch, BxBatchCommand, BxQueryBuilder, BxCallableMixin,
                                        BxQueryResponse)
from .exceptions.code_exceptions import QueryLimitExceededException, InvalidTokenException, NotFoundException
from .exceptions.bx_exceptions import EntityNotFoundException


class BaseBxRepository(object):
    GET, UPDATE, LIST, CREATE, DELETE = 'GET', 'UPDATE', 'LIST', 'CREATE', 'DELETE'
    REPOSITORY_ACTIONS = {
        GET: 'get_action',
        UPDATE: 'update_action',
        CREATE: 'create_action',
        LIST: 'list_action',
        DELETE: 'delete_action'
    }
    actions = REPOSITORY_ACTIONS

    def __init__(self, entity_cls, entity_meta, domain=settings.default_domain, token_storage=token_storage):
        # type: (BxEntity,dict, str, DefaultTokenStorage) -> None
        self.token_storage = token_storage
        self.domain = domain
        self.entity_cls = entity_cls
        self.entity_meta = entity_meta
        self.adapter = entity_meta['adapter'](entity_cls)
        super(BaseBxRepository, self).__init__()

    def _build_query_name(self, action):
        entity_name = self.entity_meta['entity']
        action = self.entity_meta[self.actions[action]]
        if action:
            action = '.' + action
        return entity_name + action

    @property
    def token(self):
        return self.token_storage.get_token(self.domain)

    def get_by_id(self, entity_id):
        # type: (str or int) -> BxEntity
        """
        Searches for entity by provided id
        :param entity_id: entity id
        :return: entity
        """
        q = BxQuery(self._build_query_name(self.GET), {'ID': entity_id, 'auth': self.token}, self.domain)
        result = self._call(q)
        return self.adapter.from_bitrix(result)

    def update(self, entity):
        # type: (BxEntity) -> BxQueryResponse
        """
        Updates entity
        :param entity: entity to update
        :return: Bitrix response
        """
        params = self.adapter.to_bitrix(entity)
        params.update({'auth': self.token})
        q = BxQuery(self._build_query_name(self.UPDATE), params, self.domain)
        result = self._call(q)
        return result

    def create(self, entity):
        # type: (BxEntity) -> BxQueryResponse
        """
        Creates entity
        :param entity:  entity to create
        :return: Bitrix response
        """
        params = self.adapter.to_bitrix(entity, True)
        params.update({'auth': self.token})
        q = BxQuery(self._build_query_name(self.CREATE), params, self.domain)
        result = self._call(q)
        return result

    def delete(self, entity):
        q = BxQuery(self._build_query_name(self.DELETE), entity.id.to_bitrix, self.domain)
        result = self._call(q)
        return result

    def get_many(self, entity_ids):
        self._validate_before_many(entity_ids, (int, str))
        if len(entity_ids) > 50:
            raise NotImplementedError('Feature for fetching more than 50 items at time not realised')
        if len(entity_ids) == 0:
            return []
        queries = [BxBatchCommand(self._build_query_name(self.GET), {'ID': i}, 'id_{}'.format(i)) for i in entity_ids]
        q = BxBatch(queries, self.token, self.domain)
        q_result = self._call(q)
        return [self.adapter.from_bitrix(bx_response) for bx_response in q_result]

    def update_many(self, entity_list):
        # type: (list or tuple or set or frozenset) -> BxBatchResponse
        """
        Updates list of entitys in one batch
        :param entity_list: list of entitys to update
        :return: Bitrix response
        """
        self._validate_before_many(entity_list, self.entity_cls)
        queries = [BxBatchCommand(self._build_query_name(self.UPDATE), entity.get_updates())
                   for entity in entity_list if entity.has_updates]
        batch = BxBatch(queries, self.token, self.domain)
        result = self._call(batch)
        return result

    @property
    def objects(self):
        # type: () -> BxEntityQuery
        return BxEntityQuery(self.entity_cls, self.entity_meta, token_storage=self.token_storage, domain=self.domain)

    @staticmethod
    def _validate_before_many(target_list, target_type):
        if not type(target_type) in (set, frozenset, list, tuple):
            target_type = (target_type,)
        if type(target_list) not in (set, frozenset, list, tuple):
            raise ValueError('Expected iterable, got {}'.format(type(target_list)))
        if not all(type(i) in target_type for i in target_list):
            raise ValueError('List members must satisfy this types {}'.format(target_type))

    def _call(self, query, token_updated=False):
        # type: (BxQuery or BxBatch) -> BxQueryResponse
        try:
            result = query.call(1)
        except QueryLimitExceededException as e:
            raise e
        except InvalidTokenException as e:
            if not token_updated:
                query.parameters.update({'auth': self.token_storage.refresh_token(self.domain)})
                result = self._call(query, True)
            else:
                raise e
        except NotFoundException as e:
            raise EntityNotFoundException(e.error, e.error_description)
        except Exception:
            raise
        return result


class BxEntityQuery(BxQueryBuilder, BxCallableMixin):
    def __init__(self,
                 target_entity_cls,
                 entity_meta,
                 token_storage=token_storage,
                 domain=settings.default_domain,
                 transport=settings.default_transport):
        # type: (BxEntity, object, str, str) -> None
        self.entity_cls = target_entity_cls
        self.token_storage = token_storage
        self.entity_meta = entity_meta
        self.adapter = entity_meta['adapter'](target_entity_cls)
        self.domain = domain
        self.transport = transport
        action = '.' + entity_meta['list_action'] if entity_meta.get('list_action') else ''
        self.url = 'https://{d}.bitrix24.ru/rest/{e}{a}.{t}'.format(d=domain,
                                                                    e=entity_meta['entity'],
                                                                    a=action,
                                                                    t=transport)
        super(BxEntityQuery, self).__init__()

    @property
    def token(self):
        return self.token_storage.get_token(self.domain)

    def __iter__(self):
        return iter(self.all())

    def resolve_field_name(self, option):
        if option in self.entity_cls.to_instance_dict:
            return self.entity_cls.to_instance_dict[option].bx_name
        else:
            return option

    def call(self, max_retries=0):
        max_retries -= 1
        try:
            params = dict(self.build(), auth=self.token)
            result = self._call(self.url, params)
        except QueryLimitExceededException:
            if max_retries >= 0:
                result = self.call(max_retries)
            else:
                raise
        return BxQueryResponse(result)

    def all(self):
        return self.adapter.from_bitrix(self.call())



# -*- coding: utf-8 -*-


class BaseBxException(Exception):
    def __init__(self, error, error_description, *args, **kwargs):
        self.error = error
        self.error_description = error_description
        for k in kwargs:
            v = kwargs[k]
            setattr(self, k, v)
        super(BaseBxException, self).__init__(*args)


class EntityNotFoundException(BaseBxException):
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from unittest import TestCase
from datetime import datetime

from bx24_orm.core.bx_interface import BxQuery, BxBatch, BxBatchCommand, BxQueryBuilder
from bx24_orm.core.bases import BxEntity
from bx24_orm.core.fields import BxField, BxDateTime
from bx24_orm.core import settings, token_storage
from bx24_orm.core.exceptions.code_exceptions import (TokenNotFoundException, InvalidQueryException,
                                                      InvalidTokenException)
from bx24_orm.enitiy.crm import BxDeal, BxLead, BxCompany

GLOBAL_TEST_LEAD = settings.TEST_LEAD
GLOBAL_TEST_DEAL = settings.TEST_DEAL
GLOBAL_TEST_COMPANY = settings.TEST_COMPANY
GLOBAL_TEST_INVOICE = settings.TEST_INVOICE


class BxQueryBuilderTest(TestCase):
    def setUp(self):
        self.TEST_IDS = [1, 2]
        self.TEST_NAME = 'TE'
        self.TEST_NUM = 35

    def testQuery(self):
        builder = BxQueryBuilder()
        expected_dict = {
            'FILTER[ID][0]': self.TEST_IDS[0],
            'FILTER[ID][1]': self.TEST_IDS[1],
            'FILTER[%TITLE]': self.TEST_NAME,
            'FILTER[>=ID]': self.TEST_NUM,
            'FILTER[<=ID]': self.TEST_NUM,
            'FILTER[>ID]': self.TEST_NUM,
            'FILTER[<ID]': self.TEST_NUM,
            'FILTER[SOME_FIELD]': self.TEST_NUM,
            'SELECT[0]': 'ID',
            'SELECT[1]': 'TITLE',
            'ORDER[ID]': 'DESC',
            'ORDER[TITLE]': 'ASC'
        }
        builder = builder.filter(id__in=self.TEST_IDS)
        builder = builder.filter(title__contains=self.TEST_NAME)
        builder = builder.filter(id__gte=self.TEST_NUM)
        builder = builder.filter(id__lte=self.TEST_NUM, id__lt=self.TEST_NUM, id__gt=self.TEST_NUM)
        builder = builder.filter(some_field=self.TEST_NUM)
        builder = builder.select('id', 'title')
        builder = builder.order('-id').order('title')
        self.assertEqual(expected_dict, builder.build())

    def testErrors(self):
        builder = BxQueryBuilder()
        self.assertRaises(ValueError, builder.filter, **{'gte': 'qwe'})
        self.assertRaises(ValueError, builder.filter, **{'id__invalid_option': 'qwe'})
        self.assertRaises(ValueError, builder.filter, **{'id__value__gte': 'qwe'})
        self.assertRaises(ValueError, builder.filter, **{'id__in': '1,2,3,4'})


class BxBatchTest(TestCase):
    def setUp(self):
        self.token = token_storage.get_token(settings.default_domain)
        self.domain = settings.default_domain
        self.TEST_LEAD = GLOBAL_TEST_LEAD
        self.TEST_DEAL = GLOBAL_TEST_DEAL
        self.TEST_COMPANY = GLOBAL_TEST_COMPANY
        self.test_queries = [('crm.lead.get', {'ID': self.TEST_LEAD}, 'raw_lead'),
                             ('crm.deal.get', {'ID': self.TEST_DEAL}, 'raw_deal'),
                             ('crm.company.get', {'ID': self.TEST_COMPANY}, 'raw_company')]

    def testConstructorFromBxBatchCommand(self):
        batches = []
        expected_batches = {}
        for q, p, n in self.test_queries:
            expected_batches.update({
                'cmd[{}]'.format(n): '{}?ID={}'.format(q, p['ID'])
            })
            batches.append(BxBatchCommand(q, p, n))
        batch = BxBatch(batches, self.token, self.domain)
        created_commands = batch.commands
        for e in expected_batches:
            self.assertEqual(created_commands[e], expected_batches[e])

    def testConstructorFromBxQueryCommand(self):
        queries = []
        expected_queries = {}
        for q, p, n in self.test_queries:
            expected_queries.update({
                'cmd[{}]'.format(n): '{}?ID={}'.format(q, p['ID'])
            })
            queries.append(BxQuery(q, p, cmd_name=n, domain=self.domain))
        batch = BxBatch(queries, self.token, self.domain)
        created_commands = batch.commands
        for e in expected_queries:
            self.assertEqual(created_commands[e], expected_queries[e])

    def testNotAsciiParameters(self):
        defaut_domain = settings.default_domain
        token = token_storage.get_token(defaut_domain)
        command = BxQuery('crm.lead.list', {'FILTER[%NAME]': 'неаскистрока', 'auth': token}, defaut_domain)
        result = command.call()
        self.assertEqual(result.total, 0)

    def testConstructorFail(self):
        expected_queries = {}
        queries = []
        for q, p, n in self.test_queries:
            expected_queries.update({
                'cmd[{}]'.format(n): '{}?ID={}'.format(q, p['ID'])
            })
            queries.append(BxQuery(q, p, cmd_name=n, domain=self.domain))
        invalid_query = {'crm.lead.get': {'ID': 1234}}
        queries.append(invalid_query)
        self.assertRaises(ValueError, BxBatch, queries, self.token, self.domain)

    def testSuccessfulQuery(self):
        batches = []
        for q, p, n in self.test_queries:
            batches.append(BxBatchCommand(q, p, n))
        batch = BxBatch(batches, self.token, self.domain)
        result = batch.call()
        for q, p, n in self.test_queries:
            self.assertEqual(result.result[n]['ID'], str(p['ID']))
            self.assertEqual(result[n].result['ID'], str(p['ID']))
            self.assertFalse(result[n].error)
            self.assertEqual(result[n].total, 1)
            self.assertEqual(result[n].result_next, 0)

    def testFailedQuery(self):
        batches = []
        for q, p, n in self.test_queries:
            batches.append(BxBatchCommand(q, p, n))
        batch = BxBatch(batches, self.token + 'INVALID_TOKEN', self.domain)
        result = batch.safe_call()
        self.assertTrue(result.error)
        self.assertFalse(result.result)
        self.assertFalse(result.result_next)
        self.assertFalse(result.result_total)


class BxQueryTest(TestCase):
    def setUp(self):
        self.TEST_LEAD_ID = GLOBAL_TEST_LEAD
        self.TEST_LEAD_QUERY = 'crm.lead.get'
        self.domain = settings.default_domain
        self.token = token_storage.get_token(settings.default_domain)

    def testTokenNotProvided(self):
        params = {'ID': self.TEST_LEAD_ID}
        query = BxQuery(self.TEST_LEAD_QUERY, params, self.domain)
        self.assertRaises(TokenNotFoundException, query.call)

    def testSuccessfulQuery(self):
        params = {'ID': self.TEST_LEAD_ID, 'auth': self.token}
        query = BxQuery(self.TEST_LEAD_QUERY, params, self.domain)
        self.assertEqual(query.call().result['ID'], self.TEST_LEAD_ID)

    def testInvalidQuery(self):
        params = {'ID': self.TEST_LEAD_ID, 'auth': self.token}
        query = BxQuery(self.TEST_LEAD_QUERY + 'MESSED_UP_QUERY', params, self.domain)
        self.assertRaises(InvalidQueryException, query.call)

    def testInvalidToken(self):
        params = {'ID': self.TEST_LEAD_ID, 'auth': self.token + 'INVALID_TOKEN'}
        query = BxQuery(self.TEST_LEAD_QUERY, params, self.domain)
        self.assertRaises(InvalidTokenException, query.call)

    def testSafeCall(self):
        params = {'ID': self.TEST_LEAD_ID, 'auth': self.token + 'INVALID_TOKEN'}
        query = BxQuery(self.TEST_LEAD_QUERY, params, self.domain)
        result = query.safe_call()
        self.assertEqual(result.result, {})
        self.assertEqual(result.result_next, 0)
        self.assertEqual(result.total, 0)
        self.assertTrue(result.error['error'])
        params = {'ID': self.TEST_LEAD_ID, 'auth': self.token}
        query = BxQuery(self.TEST_LEAD_QUERY, params, self.domain)
        result = query.safe_call()
        self.assertEqual(result.result['ID'], self.TEST_LEAD_ID)
        self.assertEqual(result.result_next, 0)
        self.assertEqual(result.total, 1)

    def testAsBatch(self):
        params = {'ID': self.TEST_LEAD_ID, 'auth': self.token}
        cmd_name = 'LEAD_GET_COMMAND'
        expected = {
            'cmd[{}]'.format(cmd_name): '{query}?ID={ID}&auth={auth}'.format(query=self.TEST_LEAD_QUERY, **params)}
        query = BxQuery(self.TEST_LEAD_QUERY, params, self.domain, cmd_name=cmd_name)
        self.assertEqual(expected, query.as_batch())


class BasesAndMixinsTest(TestCase):
    def setUp(self):
        self.TEST_ID = 1
        self.SECOND_TEST_ID = 2
        self.PHONE = [{'TYPE': 'WORK', 'VALUE': ['1234', '456']}, {'TYPE': 'HOME', 'VALUE': '4567'}]
        self.CONVERTED_COMMAND = {
            'FIELDS[PHONE][0][TYPE]': 'WORK',
            'FIELDS[PHONE][0][VALUE][0]': '1234',
            'FIELDS[PHONE][0][VALUE][1]': '456',
            'FIELDS[PHONE][1][TYPE]': 'HOME',
            'FIELDS[PHONE][1][VALUE]': '4567'
        }

    def testBxFields(self):
        now = datetime.now()
        dtime_format = '%Y-%m-%dT%H:%M:%S.%f'

        class WithBxFields(BxEntity):
            _bx_meta = {
                'default_prefix': 'FIELDS'
            }
            id = BxField('ID', prefix='FIELDS')
            phone = BxField('PHONE', prefix='FIELDS')
            dtime = BxDateTime('DATETIME')

        dealInstance = BxDeal()
        wbf = WithBxFields(ID=self.TEST_ID, PHONE=self.PHONE, DATETIME=now.strftime(dtime_format), prefix='FIELDS')
        self.assertNotEqual(dealInstance._bx_meta, wbf._bx_meta)
        self.assertFalse(wbf.changed_fields)
        self.assertEqual(wbf.id.value, self.TEST_ID)
        wbf.id = self.SECOND_TEST_ID
        self.assertEqual(wbf.id.value, self.SECOND_TEST_ID)
        self.assertEqual(wbf.changed_fields[0], 'id')
        self.assertEqual(wbf.phone.to_bitrix, self.CONVERTED_COMMAND)
        wbf.phone.value[0] = 'CHANGED'
        self.assertNotEqual(wbf.phone.value[0], 'CHANGED')
        to_change = wbf.phone.value
        to_change[0] = 'CHANGED'
        wbf.phone = to_change
        self.assertEqual(wbf.phone.value[0], 'CHANGED')
        self.assertEqual(wbf.changed_fields[1], 'phone')
        self.assertEqual(wbf.dtime.to_bitrix, {'FIELDS[DATETIME]': now.strftime(dtime_format)})
        self.assertNotEqual(wbf.phone.value, wbf.id.value)
        self.assertRaises(ValueError, wbf.__setattr__, 'dtime', 'random value')
        self.assertEqual(len(wbf.changed_fields), 2)
        wbf.dtime = now.strftime('%Y-%m-%dT%H:%M:%S.%f')
        self.assertEqual(wbf.changed_fields[2], 'dtime')
        self.assertEqual(wbf.dtime.value, now)
        self.assertEqual(len(wbf.changed_fields), 3)
        wbf.dtime = now
        self.assertEqual(len(wbf.changed_fields), 3)
        self.assertEqual(wbf.dtime.value, now)
        self.assertEqual(wbf.dtime.to_bitrix, {'FIELDS[DATETIME]': now.strftime('%Y-%m-%dT%H:%M:%S.%f')})


class BaseEntityCRUDTestMixin(object):

    def testCreateAndUpdate(self):
        import random
        import string
        before_create = len(self.entity_cls.objects.all())
        initial_entity = self.entity_cls.get(self.entity_id)
        entity = self.entity_cls(**self.entity_kwargs)
        entity.save()
        initial_entity = self.entity_cls.get(self.entity_id)
        self.assertNotEqual(entity.title(), initial_entity.title())
        created_entity = self.entity_cls.get(entity.id())
        BaseEntityCRUDTestMixin.created_entity_id = entity.id()
        self.assertNotEqual(created_entity.id(), initial_entity.id())
        for k in self.entity_kwargs:
            v = self.entity_kwargs[k]
            self.assertEqual(v, getattr(created_entity, k)())
        after_create = len(self.entity_cls.objects.all())
        self.assertNotEqual(before_create, after_create)
        self.assertEqual(after_create, before_create + 1)
        created_entity = self.entity_cls.get(BaseEntityCRUDTestMixin.created_entity_id)
        first_kwarg = next(iter(self.entity_kwargs))
        random_line = ''.join(random.choice(string.ascii_lowercase) for _ in range(10))
        setattr(created_entity, first_kwarg, random_line)
        created_entity.save()
        test_created_entity = self.entity_cls.get(BaseEntityCRUDTestMixin.created_entity_id)
        self.assertEqual(test_created_entity.id(), created_entity.id())
        self.assertEqual(getattr(test_created_entity, first_kwarg)(), random_line)

    def testDelete(self):
        before_delete = len(self.entity_cls.objects.all())
        created_entity = self.entity_cls.get(BaseEntityCRUDTestMixin.created_entity_id)
        created_entity.delete()
        after_delete = len(self.entity_cls.objects.all())
        self.assertNotEqual(before_delete, after_delete)
        self.assertEqual(before_delete, after_delete + 1)


class BxLeadTests(TestCase, BaseEntityCRUDTestMixin):
    def setUp(self):
        self.entity_id = GLOBAL_TEST_LEAD
        self.entity_cls = BxLead
        self.entity_kwargs = {'title': 'NEW_TEST_LEAD', 'utm_term': 'NEW_UTM_TERM'}

    def testCreateAndUpdate(self):
        super(BxLeadTests, self).testCreateAndUpdate()

    def testDelete(self):
        super(BxLeadTests, self).testDelete()


class BxDealTests(TestCase, BaseEntityCRUDTestMixin):
    def setUp(self):
        self.entity_id = GLOBAL_TEST_DEAL
        self.entity_cls = BxDeal
        self.entity_kwargs = {'title': 'NEW_TEST_DEAL', 'utm_term': 'NEW_UTM_TERM'}

    def testCreateAndUpdate(self):
        super(BxDealTests, self).testCreateAndUpdate()

    def testDelete(self):
        super(BxDealTests, self).testDelete()


class BxCompanyTests(TestCase, BaseEntityCRUDTestMixin):
    def setUp(self):
        self.entity_id = GLOBAL_TEST_COMPANY
        self.entity_cls = BxCompany
        self.entity_kwargs = {'title': 'NEW_TEST_COMPANY', 'utm_term': 'NEW_UTM_TERM'}

    def testCreateAndUpdate(self):
        super(BxCompanyTests, self).testCreateAndUpdate()

    def testDelete(self):
        super(BxCompanyTests, self).testDelete()


# class BxInvoiceTests(TestCase, BaseEntityCRUDTestMixin):
#     def setUp(self):
#         self.entity_id = GLOBAL_TEST_INVOICE
#         self.entity_cls = BxInvoice
#         self.entity_kwargs = {'order_topic': 'NEW_TEST_INVOICE'}
#
#     def testCreateAndUpdate(self):
#         super(BxInvoiceTests, self).testCreateAndUpdate()
#
#     def testDelete(self):
#         super(BxInvoiceTests, self).testDelete()


class DealCompanyLeadCombinationTest(TestCase):
    def setUp(self):
        self.lead_id = GLOBAL_TEST_LEAD
        self.deal_id = GLOBAL_TEST_DEAL
        self.company_id = GLOBAL_TEST_COMPANY

    def testIntegration(self):
        lead = BxLead.get(self.lead_id)
        deal = BxDeal.get(self.deal_id)
        company = BxCompany.get(self.company_id)
        self.assertEqual(company.lead_id(), lead.id())
        self.assertEqual(deal.lead_id(), lead.id())
        self.assertEqual(deal.company_id(), company.id())
        self.assertEqual(lead.company_id(), company.id())


class Deal(BxDeal):
    custom_field = BxField('UF_CRM_1539088441')


class Lead(BxLead):
    custom_field = BxField('UF_CRM_1539088367')


class Company(BxCompany):
    custom_field = BxField('UF_CRM_1539088478')


class TestLead(TestCase, BaseEntityCRUDTestMixin):
    def setUp(self):
        self.entity_id = GLOBAL_TEST_LEAD
        self.entity_cls = Lead
        self.entity_kwargs = {'title': 'NEW_TEST_LEAD', 'utm_term': 'UTM_TERM'}

    def testCreateAndUpdate(self):
        super(TestLead, self).testCreateAndUpdate()

    def testDelete(self):
        super(TestLead, self).testDelete()



# -*- coding: utf-8 -*-
import six
from copy import deepcopy

from .utils import classproperty
from .fields import BxField
from . import settings as bx_settings
from .repository import BaseBxRepository
from .adapter import BaseBxAdapter


class BxEntityMeta(type):
    _bx_meta = {
        'repository': BaseBxRepository,
        'adapter': BaseBxAdapter,
        'domain': bx_settings.default_domain,
        'transport': bx_settings.default_transport,
        'get_action': 'get',
        'delete_action': 'delete',
        'list_action': 'list',
        'update_action': 'update',
        'create_action': 'add',
        'default_prefix': ''
    }
    repository = None

    def __new__(mcs, name, bases, attrs):
        to_pop = {}
        for k in attrs:
            if issubclass(type(attrs[k]), BxField):
                to_pop.update({k: attrs[k]})
        for k in to_pop:
            attrs.pop(k)
        attrs.update({'to_instance_dict': to_pop})
        instance_bx_meta = deepcopy(mcs._bx_meta)
        instance_bx_meta.update(attrs.get('_bx_meta', {}))
        domain = instance_bx_meta['domain']
        for b in bases:
            if issubclass(b, BxEntity) and b is not BxEntity:
                base_bx_meta = deepcopy(getattr(b, '_bx_meta', {}))
                instance_bx_meta.update(base_bx_meta)
                intance_dict = deepcopy(getattr(b, 'to_instance_dict', {}))
                attrs['to_instance_dict'].update(intance_dict)
        attrs['to_instance_dict'].update({'_bx_meta': instance_bx_meta})
        super_obj = super(BxEntityMeta, mcs).__new__(mcs, name, bases, attrs)
        super_obj.repository = instance_bx_meta['repository'](super_obj,
                                                              instance_bx_meta,
                                                              domain)
        return super_obj


class BxEntity(six.with_metaclass(BxEntityMeta)):
    def __init__(self, **kwargs):
        # type: (kwargs) -> None
        provided_id = kwargs.get('ID') or None
        if (issubclass(type(provided_id), BxField)):
            id_field = provided_id
        else:
            id_field = BxField('ID', provided_id, '')
        self._bx_meta = self.to_instance_dict['_bx_meta']
        self.to_instance_dict.update({'id': id_field})
        self.to_changed_fields = []
        for k in self.to_instance_dict:
            v = self.to_instance_dict[k]
            val_type = type(v)
            if issubclass(val_type, BxField):
                if v.bx_name in kwargs:
                    value = kwargs[v.bx_name]
                elif k in kwargs:
                    value = kwargs[k]
                    self.to_changed_fields.append(k)
                else:
                    value = v.value
                if v.prefix is not None:
                    prefix = v.prefix
                else:
                    prefix = self._bx_meta['default_prefix']
                new_v = val_type(v.bx_name, value, prefix)
                setattr(self, k, new_v)
        self.changed_fields = [] + self.to_changed_fields

    def __setattr__(self, key, value):
        if key != 'changed_fields' and key in self.__dict__:
            field = getattr(self, key, None)
            if issubclass(type(field), BxField):
                field.validate_value(value)
                field.__set__(None, field.parse_value(value))
            if key not in self.changed_fields:
                self.changed_fields.append(key)
        else:
            self.__dict__.update({key: value})

    @classmethod
    def get(cls, entity_id):
        return cls.repository.get_by_id(entity_id)

    @classmethod
    def get_many(cls, entity_ids):
        return cls.repository.get_many(entity_ids)

    def save(self):
        if self.id.value:
            result = self.repository.update(self)
            is_ok = result.result
        else:
            result = self.repository.create(self)
            self.id = result.result
            is_ok = True
        self.changed_fields = []
        return self.id.value, is_ok

    def delete(self):
        self.repository.delete(self)

    @classproperty
    def objects(cls):
        # type: () -> BxEntityQuery
        return cls.repository.objects

    def get_updates(self, force_all_fields=False):
        result = {}
        if self.has_updates and not force_all_fields:
            entity_id = getattr(self, 'id')
            if not isinstance(entity_id, BxField):
                raise ValueError('id field should have BxField type')
            if entity_id.value:
                result.update(entity_id.to_bitrix)
        for f in self.changed_fields:
            field = getattr(self, f)
            if issubclass(type(field), BxField):
                result.update(field.to_bitrix)
        else:
            for f in self.__dict__:
                field = getattr(self, f)
                if issubclass(type(field), BxField):
                    result.update(field.to_bitrix)
        return result

    @property
    def has_updates(self):
        return len(self.changed_fields) > 0



# -*- coding: utf-8 -*-

import requests
import time
import uuid
from itertools import chain
from collections import OrderedDict

from six.moves import urllib
from .exceptions import code_exceptions as ce


class BxQueryBuilder(object):
    query_options = {'gte': '>=', 'gt': '>', 'lte': '<=',
                     'lt': '<', 'contains': '%', 'in': ''}

    def __init__(self):
        self._filter = {}
        self._select = []
        self._order = {}
        super(BxQueryBuilder, self).__init__()

    def resolve_field_name(self, option):
        return option

    def filter(self, **kwargs):
        if len(kwargs) == 0:
            return self
        for k in kwargs:
            v = kwargs[k]
            options = k.lower().split('__')
            options[0] = self.resolve_field_name(options[0])
            if len(options) > 2:
                raise ValueError('Invalid query: {}'.format(k))
            if options[0] in self.query_options:
                raise ValueError('Filter command should '
                                 'start with target field name. '
                                 'Found: {}'.format(options[0]))
            if len(options) == 1:
                formatted_key = 'FILTER[{}]'.format(str(options[0]).upper())
                self._filter.update({formatted_key: v})
            else:
                if options[1] not in self.query_options:
                    raise ValueError('Unexpected query option. '
                                     'Got {}. Expected {}'
                                     .format(options[1],
                                             self.query_options.keys()))
                if options[1] == 'in':
                    if type(v) not in (set, frozenset, tuple, list):
                        raise ValueError(
                            'Expected list type. Got {}'.format(type(v)))
                    for i, val in enumerate(v):
                        formatted_key = 'FILTER[{}][{}]'.format(
                            options[0].upper(), i)
                        self._filter.update({formatted_key: val})
                else:
                    formatted_key = 'FILTER[{}{}]'.format(
                        self.query_options[options[1]],
                        str(options[0]).upper())
                    self._filter.update({formatted_key: v})
        return self

    def select(self, *args):
        for a in args:
            a = self.resolve_field_name(str(a).lower())
            self._select.append(a.upper())
            # prevent identical elements in list
            # OrderedDict is used due python set swaps items
            self._select = list(OrderedDict.fromkeys(self._select))
        return self

    def build(self):
        params = {}
        params.update(self._filter)
        for k in self._order:
            v = self._order[k]
            params.update({'ORDER[{}]'.format(k): v})
        for i, v in enumerate(self._select):
            params.update({'SELECT[{}]'.format(i): v})
        return params

    def order(self, *args):
        for a in args:
            a = self.resolve_field_name(str(a).lower())
            a = str(a).upper()
            if a.startswith('-'):
                self._order.update({a[1:]: 'DESC'})
            else:
                self._order.update({a: 'ASC'})
        return self


class BxResponse(object):
    def __init__(self):
        pass

    def __parse_raw_result(self, raw_result):
        pass


class BxQueryResponse(BxResponse):
    def __init__(self, raw_result):
        # type: (dict) -> None
        """
        Represents bitrix response
        :param raw_result: response from bitrix as dict(json)
        """
        super(BxQueryResponse, self).__init__()
        self.__parse_raw_result(raw_result)

    def __parse_raw_result(self, raw_result):
        # type: (dict) -> None
        self.result = raw_result.get('result', {})
        self.error = raw_result.get('error')
        self.result_next = raw_result.get('next') or 0
        self.total = raw_result.get('total') or 1 if self.result else 0


class BxBatchResponse(BxResponse):
    def __init__(self, raw_result):
        # type: (dict) -> None
        """
        Represents batch response
        :param raw_result: response from bitrix batch as dict(json)
        """
        super(BxBatchResponse, self).__init__()
        self.__parse_raw_result(raw_result)

    def __parse_raw_result(self, raw_result):
        # type: (dict) -> None
        self.result, self.error, self.result_next, self.result_total = {}, {}, {}, {}
        self.keys = []
        if 'result' in raw_result:
            self.result = raw_result['result'].get('result') or {}
            self.error = raw_result['result'].get('result_error') or {}
            self.result_next = raw_result['result'].get('result_next') or {}
            self.result_total = raw_result['result'].get('result_total') or {}
            self.keys = chain(self.result, self.error)
        else:
            self.error = raw_result['error']

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        for k in self.keys:
            return self.__getitem__(k)
        raise StopIteration

    @property
    def cmd_names(self):
        return self.keys[:]

    def __getitem__(self, item):
        # type: (str) -> BxQueryResponse
        """
        Gets batch`s result by cmd name
        :param item: cmd name
        """
        return BxQueryResponse({'result': self.result.get(item),
                                'error': self.error.get(item),
                                'next': self.result_next.get(item) or 0,
                                'total': self.result_total.get(item)} or 0)


class BxCallableMixin(object):
    _http_code_to_exception = {
        400: ce.NotFoundException,
        401: ce.TokenExpiredException,
        403: ce.TokenNotFoundException,
        404: ce.InvalidQueryException
    }

    def __init__(self):
        super(BxCallableMixin, self).__init__()

    def call(self, max_retries):
        pass

    def safe_call(self):
        pass

    def _call(self, request_url, data):  # noqa: C901
        # type: () -> dict
        """
        Performs request, if issues error throws exception
        :return: raw result of request as dict
        """
        exception_info = {'error': '', 'error_description': ''}
        try:
            response = requests.post(request_url, data=data)
            result = response.json()
        except ValueError:
            exception_info['error'] = ce.NOT_JSON_RESPONSE
            exception_info['error_description'] = response.text
            raise ce.NotJsonResponseException(**exception_info)
        except requests.ConnectionError as e:
            exception_info['error'] = ce.NETWORK_ERROR
            exception_info[
                'error_description'] = 'Request failed due network error: {}'.format(
                str(e))
            raise ce.RequestFailedException(**exception_info)
        if 200 <= response.status_code < 300:
            result = result
        if 300 <= response.status_code < 400:
            exception_info['error'] = ce.UNEXPECTED_HTTP_CODE
            exception_info['error_description'] = response.text
            raise ce.UnexpectedResponseCodeException(**exception_info)
        elif response.status_code in (400, 401, 403, 404):
            exception_info.update(result)
            status_code = response.status_code
            if exception_info['error'] == ce.NO_AUTH_FOUND:
                status_code = 403
            raise self._http_code_to_exception[status_code](**exception_info)
        elif 400 <= response.status_code < 500:
            exception_info['error'] = 'Failed with status code {}'.format(
                response.status_code)
            exception_info['error_description'] = response.text
            raise ce.InvalidRequestException(**exception_info)
        elif response.status_code >= 500:
            exception_info = result
            if exception_info['error'] == ce.QUERY_LIMIT_EXCEEDED:
                raise ce.QueryLimitExceededException(**exception_info)
            else:
                raise ce.RequestFailedException(**exception_info)
        return result


class BxBatchCommand(object):

    def __init__(self, query, parameters, name=None):
        # type: (str, dict, str) -> None
        """
        Container for batch commands
        :param query: query like crm.deal.get and etc.
        :param parameters: query parameters
        :param name: cmd[{{name}}]
        """
        self.name = name
        self.query = query
        self.parameters = parameters

    def as_batch(self):
        # type: () -> dict
        """
        Creates batch string
        :return: dict, like { cmd[cmd_name]: url_with_params }
        """
        cmd_name = self.name if self.name else str(uuid.uuid4())[:4]
        urlencode = urllib.parse.urlencode
        return {'cmd[{}]'.format(cmd_name): '{u}?{p}'.format(u=self.query,
                                                             p=urlencode(
                                                                 self.parameters))}


class BxQuery(BxBatchCommand, BxCallableMixin):
    def __init__(self, query, params, domain, transport='json', cmd_name=None):
        # type: (str, dict, str, str, str) -> None
        """
        Performs request to bitrix
        :param query: query like crm.deal.get or other
        :param params: query parameters
        :param domain: third level bitrix domain
        :param transport: response format
        :param cmd_name: (optional) specify to prevent getting random name in as_batch method
        """
        if transport == 'xml':
            raise NotImplementedError('Xml transport feature not implemented')
        super(BxQuery, self).__init__(query, params, cmd_name)
        self.transport = transport
        self.request_url = 'https://{}.bitrix24.ru/rest/{}.{}/'.format(domain,
                                                                       query,
                                                                       transport)

    def call(self, max_retries=0):
        # type: (int) -> BxQueryResponse
        """
        Performs request to bitrix with opportunity to repeat request if query limit exceeded
        :param max_retries: number of retries before give up in case of QUERY_LIMIT_EXCEEDED Exception
        :return: result of query
        """
        max_retries -= 1
        try:
            raw_result = self._call(self.request_url, self.parameters)
            result = BxQueryResponse(raw_result)
        except ce.QueryLimitExceededException:
            if max_retries >= 0:
                time.sleep(2)
                result = self.call(max_retries)
            else:
                raise
        return result

    def safe_call(self):
        # type: () -> BxQueryResponse
        """
        Same as call, but instead of raising exception send it in result
        :return: BxResponse object
        """
        try:
            return self.call(max_retries=1)
        except ce.BxException as e:
            error = {'error': e.error,
                     'error_description': e.error_description}
        except Exception:
            raise
        return BxQueryResponse({'error': error})


class BxBatch(BxCallableMixin):
    def __init__(self, commands, token, domain, halt=0, transport='json'):
        # type: (list, str, str, int, str) -> None
        """
        Creates bitrix batch request object
        :param commands: iterable of BxBatchCommand objects
        :param token: access token
        :param halt: terminate batch if one of commands failed
        :param domain: portal`s third level domain
        :param transport: response format
        """
        super(BxBatch, self).__init__()
        if transport == 'xml':
            raise NotImplementedError('Xml transport feature not implemented')
        self.request_url = 'https://{}.bitrix24.ru/rest/batch.{}/'.format(
            domain, transport)
        self.parameters = {'halt': halt, 'auth': token}
        self.commands = {}
        for c in commands:
            if not isinstance(c, BxBatchCommand):
                raise ValueError(
                    'Command {} type is not <BxBatchCommand>'.format(c))
            self.commands.update(c.as_batch())

    def call(self, max_retries=0):
        # type: (int) -> BxBatchResponse
        """
        Performs batch call
        :param max_retries: Number of tries before give in in case of QUERY_LIMIT_EXCEEDED exception
        :return:
        """
        max_retries -= 1
        try:
            data = dict(self.parameters, **self.commands)
            result = BxBatchResponse(self._call(self.request_url, data))
        except ce.QueryLimitExceededException:
            if max_retries >= 0:
                time.sleep(2)
                result = self.call(max_retries)
            else:
                raise
        return result

    def safe_call(self):
        # type: () -> BxBatchResponse
        """
        Same as call, but instead of raising exception send it in result
        :return: BxBatchResponse object
        """
        try:
            result = self.call(max_retries=1)
        except ce.BxException as e:
            result = BxBatchResponse({'error': {'error': e.error,
                                                'error_description': e.error_description}})
        except Exception:
            raise
        return result



# -*- coding: utf-8 -*-
from copy import deepcopy
from datetime import datetime
from dateutil.parser import parse
import six


class BxField(object):
    def __init__(self, bx_name, value=None, prefix=None):
        # type: (six.string_types, object, six.string_types) -> None
        """
        Represents bitrix field
        :param bx_name: name on bitrix portal like UF_CRM_XXXXXX
        :param value: initial value
        :param prefix: prefix before parameters dict, e.g FIELD in ORM module
        """
        self._value = value
        self._bx_name = bx_name.upper()
        self._prefix = prefix.upper() if issubclass(type(prefix), six.string_types) else None
        super(BxField, self).__init__()

    def parse_value(self, value):
        return value

    @property
    def prefix(self):
        return self._prefix

    @property
    def bx_name(self):
        return self._bx_name

    @property
    def value(self):
        # type: () -> object
        """
        Returns copy of contained value to prevent untracked changes
        :return: value
        """
        return deepcopy(self._value)

    def __set__(self, instance, value):
        self._value = value

    @staticmethod
    def __format_value(key, value):
        # type: (str, object or list or tuple or set or frozenset) -> dict
        """
        Recursively unwrap field to properly formed dict of params
        :param key: current key
        :param value: value to unwrap
        :return: dict of parameters
        """
        result = {}
        if type(value) in (list, tuple, set, frozenset):
            for i, v in enumerate(value):
                result.update(BxField.__format_value(key + '[{}]'.format(i), v))
        elif type(value) is dict:
            for k in value:
                v = value[k]
                result.update(BxField.__format_value(key + '[{}]'.format(k.upper()), v))
        elif value is not None:
            result.update({key: value})
        return result

    def validate_value(self, value):
        return True

    def __call__(self, *args, **kwargs):
        return deepcopy(self._value)

    def __str__(self):
        return self._value.__str__()

    @property
    def to_bitrix(self):
        # type: () -> dict
        """
        Converts field into valid parameter or parameters to bitrix
        :return: Dict representing bitrix parameter
        """
        if self._prefix:
            key = '{}[{}]'.format(self._prefix, self._bx_name)
        else:
            key = '{}'.format(self._bx_name)
        return self.__format_value(key, self.value)


class BxDateTime(BxField):

    def __init__(self, bx_name, value=None, prefix=None):
        try:
            self.validate_value(value)
            value = self.parse_value(value)
        except ValueError as err:
            raise err
        super(BxDateTime, self).__init__(bx_name, value, prefix)

    def parse_value(self, value):
        if not value:
            return None
        if type(value) == datetime:
            return value
        else:
            return parse(value)

    def validate_value(self, value):
        if isinstance(value, datetime):
            return True
        if value is None:
            return None
        if issubclass(type(value), six.string_types):
            if value.strip() == '':
                return None
            else:
                try:
                    parse(value)
                    return True
                except ValueError as err:
                    raise err
        raise ValueError('Datetime or str instance expected. Got {}'.format(type(value)))

    @property
    def to_bitrix(self):
        if self._prefix:
            key = '{}[{}]'.format(self._prefix, self._bx_name)
        else:
            key = '{}'.format(self._bx_name)
        if self._value:
            try:
                v = self._value.strftime('%Y-%m-%dT%H:%M:%S.%f%Z')
            except ValueError:
                v = self._value.strftime('%Y-%m-%dT%H:%M:%S.%f')
        else:
            v = ''
        return {key: v}


class BxBoolean(BxField):
    def __init__(self, bx_name, value=None, prefix=None):
        if self.validate_value(value):
            value = self.parse_value(value)
            super(BxBoolean, self).__init__(bx_name, value, prefix)
        else:
            raise RuntimeError('Failed to initialize BxBoolean with value = {}'.format(value))

    def parse_value(self, value):
        if value is None:
            return False
        if issubclass(type(value), six.string_types) and value.upper().strip() in ('Y', 'N'):
            return True if value.upper().strip() is 'Y' else False
        if type(value) is bool:
            return value
        raise ValueError('Expected Y/N or bool or None. Got {}'.format(type(value)))

    def validate_value(self, value):
        if value is None:
            return True
        if issubclass(type(value), six.string_types) and value.upper().strip() in ('Y', 'N'):
            return True
        if type(value) is bool:
            return True
        raise ValueError('Expected Y/N or bool or None. Got {}'.format(type(value)))

    @property
    def to_bitrix(self):
        if self._prefix:
            key = '{}[{}]'.format(self._prefix, self._bx_name)
        else:
            key = '{}'.format(self._bx_name)
        v = 'Y' if self._value else 'N'
        return {key: v}



# -*- coding: utf-8 -*-

import requests
import shelve

from os import path


class BaseTokenStorage(object):
    def __init__(self, *args, **kwargs):
        super(BaseTokenStorage, self).__init__(*args, **kwargs)

    def save_token(self, domain, token, refresh_token):
        """
        Save token and refresh token
        :param domain: your bitrix 3-d level domain
        :param token: access_token
        :param refresh_token: refresh_token
        """
        raise NotImplementedError()

    def get_token(self, domain):
        """
        Gets access token
        :rtype: str
        :param domain: your bitrix 3-d level domain
        """
        raise NotImplementedError()

    def get_refresh_token(self, domain):
        """
        Gets refresh token
        :rtype: str
        :param domain: your bitrix 3-d level domain
        """
        raise NotImplementedError()

    def refresh_token(self, domain):
        """
        Gets access token
        :rtype: str
        :param domain: your bitrix 3-d level domain
        """
        raise NotImplementedError()

    def get_client_credentials(self, domain):
        """
        Gets dict of client_id and client_secret credentials
        :rtype: dict
        :param domain: your bitrix 3-d level domain
        """
        raise NotImplementedError()


class DefaultTokenStorage(BaseTokenStorage):
    def __init__(self, bx_settings):
        self.bx_settings = bx_settings
        super(DefaultTokenStorage, self).__init__()

    def save_token(self, domain, token, refresh_token):
        storage = shelve.open(path.abspath(self.bx_settings.TOKEN_STORAGE_FILE_PATH))
        storage[domain] = {'token': token, 'refresh_token': refresh_token}
        storage.close()

    def get_client_credentials(self, domain):
        credentials = self.bx_settings.BX24_DOMAIN_SETTINGS[domain]
        return {
            'client_id': credentials['client_id'],
            'client_secret': credentials['client_secret']
        }

    def get_token(self, domain):
        # type: (str) -> str
        storage = shelve.open(self.bx_settings.TOKEN_STORAGE_FILENAME)
        result = storage[domain]['token']
        storage.close()
        return result

    def get_refresh_token(self, domain):
        # type: (str) -> str
        storage = shelve.open(self.bx_settings.TOKEN_STORAGE_FILENAME)
        result = storage[domain]['refresh_token']
        storage.close()
        return result

    def refresh_token(self, domain):
        # type: (str) -> str
        url = 'https://{}.bitrix24.ru/oauth/token/'.format(domain)
        domain_options = self.get_client_credentials(domain)
        params = {
            'client_id': domain_options['client_id'],
            'client_secret': domain_options['client_secret'],
            'grant_type': 'refresh_token',
            'refresh_token': self.get_refresh_token(domain)
        }
        response = requests.get(url, params)
        result = response.json()
        token, refresh_token = result['access_token'], result['refresh_token']
        self.save_token(domain, token, refresh_token)
        return token



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
setup(name='bx24_orm',
      version='0.0.1',
      description='Easy to use Django-styled API wrapper for Bitrix 24',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/dmitriilazukov/bx24_orm',
      author='Dmitrii Lazukov',
      author_email='dmitriilazukov@gmail.com',
      license='MIT',
      packages=find_packages(where='.', exclude='bx24_settings'),
      entry_points={
          'console_scripts': ['bx24_cmd=bx24_orm.scripts.command_line:bx24_cmd'],
      },
      install_requires=[
          'requests',
          'six',
          'python-dateutil'
      ],
      classifiers=[
          "Programming Language :: Python :: 2.7",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      zip_safe=False)



# -*- coding: utf-8 -*-


from bx24_orm.core.bases import BxEntity
from bx24_orm.core.fields import BxField, BxDateTime, BxBoolean


class BxLead(BxEntity):
    title = BxField('TITLE')
    name = BxField('NAME')
    second_name = BxField('SECOND_NAME')
    last_name = BxField('LAST_NAME')
    company_title = BxField('COMPANY_TITLE')
    source_id = BxField('SOURCE_ID')
    source_description = BxField('SOURCE_DESCRIPTION')
    status_id = BxField('STATUS_ID')
    status_description = BxField('STATUS_DESCRIPTION')
    post = BxField('POST')
    address = BxField('ADDRESS')
    address_2 = BxField('ADDRESS_2')
    address_city = BxField('ADDRESS_CITY')
    address_postal_code = BxField('ADDRESS_POSTAL_CODE')
    address_region = BxField('ADDRESS_REGION')
    address_province = BxField('ADDRESS_PROVINCE')
    address_country = BxField('ADDRESS_COUNTRY')
    address_country_code = BxField('ADDRESS_COUNTRY_CODE')
    currency_id = BxField('CURRENCY_ID')
    opportunity = BxField('OPPORTUNITY')
    opened = BxBoolean('OPENED')
    comments = BxField('COMMENTS')
    assigned_by_id = BxField('ASSIGNED_BY_ID')
    created_by_id = BxField('CREATED_BY_ID')
    modify_by_id = BxField('MODIFY_BY_ID')
    date_create = BxDateTime('DATE_CREATE')
    date_modify = BxDateTime('DATE_MODIFY')
    company_id = BxField('COMPANY_ID')
    contact_id = BxField('CONTACT_ID')
    date_closed = BxDateTime('DATE_CLOSED')
    phone = BxField('PHONE')
    email = BxField('EMAIL')
    web = BxField('WEB')
    im = BxField('IM')
    originator_id = BxField('ORIGINATOR_ID')
    origin_id = BxField('ORIGIN_ID')
    honorific = BxField('HONORIFIC')
    is_return_customer = BxBoolean('IS_RETURN_CUSTOMER')
    has_phone = BxBoolean('HAS_PHONE')
    has_email = BxBoolean('HAS_EMAIL')
    has_imol = BxBoolean('HAS_IMOL')
    status_semantic_id = BxField('STATUS_SEMANTIC_ID')
    utm_source = BxField('UTM_SOURCE')
    utm_medium = BxField('UTM_MEDIUM')
    utm_campaign = BxField('UTM_CAMPAIGN')
    utm_content = BxField('UTM_CONTENT')
    utm_term = BxField('UTM_TERM')
    _bx_meta = {
        'entity': 'crm.lead',
        'default_prefix': 'FIELDS'
    }


class BxDeal(BxEntity):
    title = BxField('TITLE')
    type_id = BxField('TYPE_ID')
    stage = BxField('STAGE_ID')
    probability = BxField('PROBABILITY')
    opportunity = BxField('OPPORTUNITY')
    company_id = BxField('COMPANY_ID')
    contact_id = BxField('CONTACT_ID')
    begin_date = BxDateTime('BEGIN_DATE')
    close_date = BxDateTime('CLOSE_DATE')
    opened = BxBoolean('OPENED')
    closed = BxBoolean('CLOSED')
    comments = BxField('COMMENTS')
    assigned_by_id = BxField('ASSIGNED_BY_ID')
    crated_by_id = BxField('CREATED_BY_ID')
    modify_by_id = BxField('MODIFY_BY_ID')
    date_create = BxDateTime('DATE_CREATE')
    date_modify = BxDateTime('DATE_MODIFY')
    lead_id = BxField('LEAD_ID')
    additional_info = BxField('ADDITIONAL_INFO')
    originator_id = BxField('ORIGINATOR_ID')
    origin_id = BxField('ORIGIN_ID')
    category_id = BxField('CATEGORY_ID')
    currency_id = BxField('CURRENCY_ID')
    tax_value = BxField('TAX_VALUE')
    quote_id = BxField('QUOTE_ID')
    location_id = BxField('LOCATION_ID')
    stage_semantic_id = BxField('STAGE_SEMANTIC_ID')
    is_new = BxBoolean('IS_NEW')
    is_recurring = BxBoolean('IS_RECURRING')
    is_return_customer = BxBoolean('IS_RETURN_CUSTOMER')
    is_repeated_approach = BxBoolean('IS_REPEATED_APPROACH')
    source_id = BxField('SOURCE_ID')
    source_description = BxField('SOURCE_DESCRIPTION')
    utm_source = BxField('UTM_SOURCE')
    utm_medium = BxField('UTM_MEDIUM')
    utm_campaign = BxField('UTM_CAMPAIGN')
    utm_content = BxField('UTM_CONTENT')
    utm_term = BxField('UTM_TERM')
    _bx_meta = {
        'entity': 'crm.deal',
        'default_prefix': 'FIELDS'
    }


class BxCompany(BxEntity):
    title = BxField('TITLE')
    company_type = BxField('COMPANY_TYPE')
    logo = BxField('LOGO')
    address = BxField('ADDRESS')
    address_2 = BxField('ADDRESS_2')
    address_city = BxField('ADDRESS_CITY')
    address_postal_code = BxField('ADDRESS_POSTAL_CODE')
    address_region = BxField('ADDRESS_REGION')
    address_province = BxField('ADDRESS_PROVINCE')
    address_country = BxField('ADDRESS_COUNTRY')
    address_country_code = BxField('ADDRESS_COUNTRY_CODE')
    reg_address = BxField('REG_ADDRESS')
    reg_address_2 = BxField('REG_ADDRESS_2')
    reg_address_city = BxField('REG_ADDRESS_CITY')
    reg_address_postal_code = BxField('REG_ADDRESS_POSTAL_CODE')
    reg_address_region = BxField('REG_ADDRESS_REGION')
    reg_address_province = BxField('REG_ADDRESS_PROVINCE')
    reg_address_country = BxField('REG_ADDRESS_COUNTRY')
    reg_address_country_code = BxField('REG_ADDRESS_COUNTRY_CODE')
    address_legal = BxField('ADDRESS_LEGAL')
    banking_details = BxField('BANKING_DETAILS')
    industry = BxField('INDUSTRY')
    currency_id = BxField('CURRENCY_ID')
    revenue = BxField('REVENUE')
    opened = BxBoolean('OPENED')
    comments = BxField('COMMENTS')
    assigned_by_id = BxField('ASSIGNED_BY_ID')
    created_by_id = BxField('CREATED_BY_ID')
    modify_by_id = BxField('MODIFY_BY_ID')
    date_create = BxDateTime('DATE_CREATE')
    date_modify = BxDateTime('DATE_MODIFY')
    lead_id = BxField('LEAD_ID')
    originator_id = BxField('ORIGINATOR_ID')
    origin_id = BxField('ORIGIN_ID')
    phone = BxField('PHONE')
    email = BxField('EMAIL')
    web = BxField('WEB')
    im = BxField('IM')
    has_phone = BxBoolean('HAS_PHONE')
    has_email = BxBoolean('HAS_EMAIL')
    has_imol = BxBoolean('HAS_IMOL')
    employees = BxField('EMPLOYEES')
    is_my_company = BxBoolean('IS_MY_COMPANY')
    origin_version = BxField('ORIGIN_VERSION')
    utm_source = BxField('UTM_SOURCE')
    utm_medium = BxField('UTM_MEDIUM')
    utm_campaign = BxField('UTM_CAMPAIGN')
    utm_content = BxField('UTM_CONTENT')
    utm_term = BxField('UTM_TERM')

    _bx_meta = {
        'entity': 'crm.company',
        'default_prefix': 'FIELDS'
    }


class BxContact(BxEntity):
    name = BxField('NAME')
    second_name = BxField('SECOND_NAME')
    last_name = BxField('LAST_NAME')
    photo = BxField('PHOTO')
    birthdate = BxField('BIRTHDATE')
    type_id = BxField('TYPE_ID')
    source_id = BxField('SOURCE_ID')
    source_description = BxField('SOURCE_DESCRIPTION')
    post = BxField('POST')
    address = BxField('ADDRESS')
    address_2 = BxField('ADDRESS_2')
    address_city = BxField('ADDRESS_CITY')
    address_postal_code = BxField('ADDRESS_POSTAL_CODE')
    address_region = BxField('ADDRESS_REGION')
    address_province = BxField('ADDRESS_PROVINCE')
    address_country = BxField('ADDRESS_COUNTRY')
    address_country_code = BxField('ADDRESS_COUNTRY_CODE')
    opened = BxBoolean('OPENED')
    comments = BxField('COMMENTS')
    export = BxBoolean('EXPORT')
    assigned_by_id = BxField('ASSIGNED_BY_ID')
    created_by_id = BxField('CREATED_BY_ID')
    modify_by_id = BxField('MODIFY_BY_ID')
    date_create = BxDateTime('DATE_CREATE')
    date_modify = BxDateTime('DATE_MODIFY')
    company_id = BxField('COMPANY_ID')
    lead_id = BxField('LEAD_ID')
    phone = BxField('PHONE')
    email = BxField('EMAIL')
    web = BxField('WEB')
    im = BxField('IM')
    originator_id = BxField('ORIGINATOR_ID')
    origin_id = BxField('ORIGIN_ID')
    honorific = BxField('HONORIFIC')
    has_phone = BxBoolean('HAS_PHONE')
    has_email = BxBoolean('HAS_EMAIL')
    has_imol = BxBoolean('HAS_IMOL')
    origin_version = BxField('ORIGIN_VERSION')
    face_id = BxField('FACE_ID')
    utm_source = BxField('UTM_SOURCE')
    utm_medium = BxField('UTM_MEDIUM')
    utm_campaign = BxField('UTM_CAMPAIGN')
    utm_content = BxField('UTM_CONTENT')
    utm_term = BxField('UTM_TERM')

    _bx_meta = {
        'entity': 'crm.contact',
        'default_prefix': 'FIELDS'
    }


class BxActivity(BxEntity):
    owner_id = BxField('OWNER_ID')
    owner_type_id = BxField('OWNER_TYPE_ID')
    type_id = BxField('TYPE_ID')
    provider_id = BxField('PROVIDER_ID')
    provider_type_id = BxField('PROVIDER_TYPE_ID')
    provider_group_id = BxField('PROVIDER_GROUP_ID')
    associated_entity_id = BxField('ASSOCIATED_ENTITY_ID')
    subject = BxField('SUBJECT')
    created = BxDateTime('CREATED')
    last_updated = BxDateTime('LAST_UPDATED')
    start_time = BxDateTime('START_TIME')
    end_time = BxDateTime('END_TIME')
    deadline = BxDateTime('DEADLINE')
    completed = BxBoolean('COMPLETED')
    status = BxField('STATUS')
    responsible_id = BxField('RESPONSIBLE_ID')
    priority = BxField('PRIORITY')
    notify_type = BxField('NOTIFY_TYPE')
    notify_value = BxField('NOTIFY_VALUE')
    description = BxField('DESCRIPTION')
    description_type = BxField('DESCRIPTION_TYPE')
    direction = BxField('DIRECTION')
    location = BxField('LOCATION')
    settings = BxField('SETTINGS')
    originator_id = BxField('ORIGINATOR_ID')
    origin_id = BxField('ORIGIN_ID')
    author_id = BxField('AUTHOR_ID')
    editor_id = BxField('EDITOR_ID')
    provider_params = BxField('PROVIDER_PARAMS')
    provider_data = BxField('PROVIDER_DATA')
    result_mark = BxField('RESULT_MARK')
    result_value = BxField('RESULT_VALUE')
    result_sum = BxField('RESULT_SUM')
    result_currency_id = BxField('RESULT_CURRENCY_ID')
    result_status = BxField('RESULT_STATUS')
    result_stream = BxField('RESULT_STREAM')
    result_source_id = BxField('RESULT_SOURCE_ID')
    autocomplete_rule = BxField('AUTOCOMPLETE_RULE')

    _bx_meta = {
        'entity': 'crm.activity',
        'default_prefix': 'FIELDS'
    }


class BxMeasure(BxEntity):
    code = BxField('CODE')
    measure_title = BxField('MEASURE_TITLE')
    symbol_rus = BxField('SYMBOL_RUS')
    symbol_intl = BxField('SYMBOL_INTL')
    symbol_letter_intl = BxField('SYMBOL_LETTER_INTL')
    is_default = BxBoolean('IS_DEFAULT')

    _bx_meta = {
        'entity': 'crm.measure',
        'default_prefix': 'FIELDS'
    }


class BxCatalog(BxEntity):
    originator_id = BxField('ORIGINATOR_ID')
    origin_id = BxField('ORIGIN_ID')
    name = BxField('NAME')
    xml_id = BxField('XML_ID')

    _bx_meta = {
        'entity': 'crm.catalog',
        'default_prefix': 'FIELDS'
    }


class BxQuote(BxEntity):
    title = BxField('ENTITY')
    status_id = BxField('STATUS_ID')
    currency_id = BxField('CURRENCY_ID')
    opportunity = BxField('OPPORTUNITY')
    tax_value = BxField('TAX_VALUE')
    company_id = BxField('COMPANY_ID')
    contact_id = BxField('CONTACT_ID')
    mycompany_id = BxField('MYCOMPANY_ID')
    begin_date = BxField('BEGINDATE')
    close_date = BxField('CLOSEDATE')
    assigned_by_id = BxField('ASSIGNED_BY_ID')
    created_by_id = BxField('CREATED_BY_ID')
    modify_by_id = BxField('MODIFY_BY_ID')
    date_create = BxField('DATE_CREATE')
    date_modify = BxField('DATE_MODIFY')
    opened = BxBoolean('OPENED')
    closed = BxBoolean('CLOSED')
    comments = BxField('COMMENTS')
    lead_id = BxField('LEAD_ID')
    deal_id = BxField('DEAL_ID')
    quote_member = BxField('QUOTE_MEMBER')
    content = BxField('CONTENT')
    terms = BxField('TERMS')
    person_type_id = BxField('PERSON_TYPE_ID')
    location_id = BxField('LOCATION_ID')
    client_title = BxField('CLIENT_TITLE')
    client_addr = BxField('CLIENT_ADDR')
    client_contact = BxField('CLIENT_CONTACT')
    client_email = BxField('CLIENT_EMAIL')
    client_phone = BxField('CLIENT_PHONE')
    client_tp_id = BxField('CLIENT_TP_ID')
    client_tpa_id = BxField('CLIENT_TPA_ID')
    utm_source = BxField('UTM_SOURCE')
    utm_medium = BxField('UTM_MEDIUM')
    utm_campaign = BxField('UTM_CAMPAIGN')
    utm_content = BxField('UTM_CONTENT')
    utm_term = BxField('UTM_TERM')

    _bx_meta = {
        'entity': 'crm.quote',
        'default_prefix': 'FIELDS'
    }


class BxDealCategoryStage(BxEntity):
    name = BxField('NAME')
    sort = BxField('SORT')
    status_id = BxField('STATUS_ID')

    _bx_meta = {
        'entity': 'crm.dealcategory.stage',
        'default_prefix': 'FIELDS'
    }


class BxProductSection(BxEntity):
    catalog_id = BxField('CATALOG_ID')
    section_id = BxField('SECTION_ID')
    name = BxField('NAME')
    xml_id = BxField('XML_ID')

    _bx_meta = {
        'entity': 'crm.productsection',
        'default_prefix': 'FIELDS'
    }


class BxAddress(BxEntity):
    type_id = BxField('ENITY_ID')
    entity_type_id = BxField('ENTITY_TYPE_ID')
    entity_id = BxField('ENTITY_ID')
    address_1 = BxField('ADDRESS_1')
    address_2 = BxField('ADDRESS_2')
    city = BxField('CITY')
    postal_code = BxField('POSTAL_CODE')
    region = BxField('REGION')
    province = BxField('PROVINCE')
    country = BxField('COUNTRY')
    country_code = BxField('COUNTRY_CODE')
    anchor_type_id = BxField('ANCHOR_TYPE_ID')
    anchor_id = BxField('ANCHOR_ID')

    _bx_meta = {
        'entity': 'crm.address',
        'default_prefix': 'FIELDS'
    }


class BxRequisite(BxEntity):
    entity_type_id = BxField('ENTITY_TYPE_ID')
    entity_id = BxField('ENTITY_ID')
    preset_id = BxField('PRESET_ID')
    date_create = BxDateTime('DATE_CREATE')
    date_modify = BxDateTime('DATE_MODIFY')
    created_by_id = BxField('CREATED_BY_ID')
    modify_by_id = BxField('MODIFY_BY_ID')
    name = BxField('NAME')
    code = BxField('CODE')
    xml_id = BxField('XML_ID')
    originator_id = BxField('ORIGINATOR_ID')
    active = BxBoolean('ACTIVE')
    sort = BxField('SORT')
    rq_name = BxField('RQ_NAME')
    rq_first_name = BxField('RQ_FIRST_NAME')
    rq_last_name = BxField('RQ_LAST_NAME')
    rq_second_name = BxField('RQ_SECOND_NAME')
    rq_company_name = BxField('RQ_COMPANY_NAME')
    rq_company_full_name = BxField('RQ_COMPANY_FULL_NAME')
    rq_company_reg_date = BxField('RQ_COMPANY_REG_DATE')
    rq_director = BxField('RQ_DIRECTOR')
    rq_accountant = BxField('RQ_ACCOUNTANT')
    rq_ceo_name = BxField('RQ_CEO_NAME')
    rq_ceo_work_pos = BxField('RQ_CEO_WORK_POS')
    rq_contact = BxField('RQ_CONTACT')
    rq_email = BxField('RQ_EMAIL')
    rq_phone = BxField('RQ_PHONE')
    rq_fax = BxField('RQ_FAX')
    rq_ident_doc = BxField('RQ_IDENT_DOC')
    rq_ident_doc_ser = BxField('RQ_IDENT_DOC_SER')
    rq_ident_doc_num = BxField('RQ_IDENT_DOC_NUM')
    rq_ident_doc_pers_num = BxField('RQ_IDENT_DOC_PERS_NUM')
    rq_ident_doc_date = BxField('RQ_IDENT_DOC_DATE')
    rq_ident_doc_issued_by = BxField('RQ_IDENT_DOC_ISSUED_BY')
    rq_ident_doc_dep_code = BxField('RQ_IDENT_DOC_DEP_CODE')
    rq_inn = BxField('RQ_INN')
    rq_kpp = BxField('RQ_KPP')
    rq_usrle = BxField('RQ_USRLE')
    rq_ifsn = BxField('RQ_IFSN')
    rq_ogrn = BxField('RQ_OGRN')
    rq_ogrnip = BxField('RQ_OGRNIP')
    rq_okpo = BxField('RQ_OKPO')
    rq_oktmo = BxField('RQ_OKTMO')
    rq_okved = BxField('RQ_OKVED')
    rq_edrpou = BxField('RQ_EDRPOU')
    rq_drfo = BxField('RQ_DRFO')
    rq_kbe = BxField('RQ_KBE')
    rq_iin = BxField('RQ_IIN')
    rq_bin = BxField('RQ_BIN')
    rq_st_cert_ser = BxField('RQ_ST_CERT_SER')
    rq_st_cert_num = BxField('RQ_ST_CERT_NUM')
    rq_st_cert_date = BxField('RQ_ST_CERT_DATE')
    rq_vat_payer = BxBoolean('RQ_VAT_PAYER')
    rq_vat_id = BxField('RQ_VAT_ID')
    rq_vat_cert_ser = BxField('RQ_VAT_CERT_SER')
    rq_vat_cert_num = BxField('RQ_VAT_CERT_NUM')
    rq_vat_cert_date = BxField('RQ_VAT_CERT_DATE')
    rq_recidence_country = BxField('RQ_RECIDENCE_COUNTRY')
    rq_base_doc = BxField('RQ_BASE_DOC')

    _bx_meta = {
        'entity': 'crm.requisite',
        'default_prefix': 'FIELDS'
    }


class BxRequisiteBankdetail(BxEntity):
    entity_id = BxField('ENTITY_ID')
    country_id = BxField('COUNTRY_ID')
    date_create = BxDateTime('DATE_CREATE')
    date_modify = BxDateTime('DATE_MODIFY')
    name = BxField('NAME')
    code = BxField('CODE')
    xml_id = BxField('XML_ID')
    active = BxBoolean('ACTIVE')
    sort = BxField('SORT')
    rq_bank_name = BxField('RQ_BANK_NAME')
    rq_bank_addr = BxField('RQ_BANK_ADDR')
    rq_bank_route_num = BxField('RQ_BANK_ROUTE_NUM')
    rq_bik = BxField('RQ_BIK')
    rq_mfo = BxField('RQ_MFO')
    rq_acc_name = BxField('RQ_ACC_NAME')
    rq_acc_num = BxField('RQ_ACC_NUM')
    rq_iik = BxField('RQ_IIK')
    rq_acc_currency = BxField('RQ_ACC_CURRENCY')
    rq_cor_acc_num = BxField('RQ_COR_ACC_NUM')
    rq_iban = BxField('RQ_IBAN')
    rq_swift = BxField('RQ_SWIFT')
    rq_bic = BxField('RQ_BIC')
    comments = BxField('COMMENTS')
    originator_id = BxField('ORIGINATOR_ID')

    _bx_meta = {
        'entity': 'crm.requisiste.bankdetail',
        'default_prefix': 'FIELDS'
    }


class BxRequisiteLink(BxEntity):
    entity_type_id = BxField('ENITY_TYPE_ID')
    entity_id = BxField('ENTITY_ID')
    requisite_id = BxField('REQUISITE_ID')
    bank_detail_id = BxField('BANK_DETAIL_ID')
    mc_requisite_id = BxField('MC_REQUISITE_ID')
    mc_bank_detail_id = BxField('MC_BANK_DETAIL_ID')

    _bx_meta = {
        'entity': 'crm.requisite.link',
        'default_prefix': 'FIELDS'
    }


class BxRequisitePreset(BxEntity):
    entity_type_id = BxField('ENTITY_TYPE_ID')
    country_id = BxField('COUNTRY_ID')
    name = BxField('NAME')
    date_create = BxDateTime('DATE_CREATE')
    date_modify = BxDateTime('DATE_MODIFY')
    created_by_id = BxField('CREATED_BY_ID')
    modify_by_id = BxField('MODIFY_BY_ID')
    active = BxBoolean('ACTIVE')
    sort = BxField('SORT')
    xml_id = BxField('XML_ID')

    _bx_meta = {
        'entity': 'crm.requisite.preset',
        'default_prefix': 'FIELDS'
    }


class BxInvoiceStatus(BxEntity):
    entity_id = BxField('ENTITY_ID')
    status_id = BxField('STATUS_ID')
    name = BxField('NAME')
    name_init = BxField('NAME_INIT')
    sort = BxField('SORT')
    system = BxBoolean('SYSTEM')

    _bx_meta = {
        'entity': 'crm.invoice.status',
        'default_prefix': 'FIELDS'
    }


class BxDealCategory(BxEntity):
    created_date = BxDateTime('CREATED_DATE')
    name = BxDateTime('NAME')
    is_locked = BxBoolean('IS_LOCKED')
    sort = BxField('SORT')

    _bx_meta = {
        'entity': 'crm.dealcategory'
    }


class BxInvoice(BxEntity):
    account_number = BxField('ACCOUNT_NUMBER')
    comments = BxField('COMMENTS')
    currency = BxField('CURRENCY')
    date_bill = BxDateTime('DATE_BILL')
    date_insert = BxDateTime('DATE_INSERT')
    date_marked = BxDateTime('DATE_MARKED')
    date_pay_before = BxDateTime('DATE_PAY_BEFORE')
    date_payed = BxDateTime('DATE_PAYED')
    date_status = BxDateTime('DATE_STATUS')
    date_update = BxDateTime('DATE_UPDATE')
    created_by = BxField('CREATED_BY')
    emp_payed_id = BxField('EMP_PAYED_ID')
    emp_status_id = BxField('EMP_STATUS_ID')
    lid = BxField('LID')
    xml_id = BxField('XML_ID')
    order_topic = BxField('ORDER_TOPIC')
    pay_system_id = BxField('PAY_SYSTEM_ID')
    pay_voucher_date = BxDateTime('PAY_VOUCHER_DATE')
    pay_voucher_num = BxField('PAY_VOUCHER_NUM')
    payed = BxBoolean('PAYED')
    person_type_id = BxField('PERSON_TYPE_ID')
    price = BxField('PRICE')
    reason_marked = BxField('REASON_MARKED')
    responsible_email = BxField('RESPONSIBLE_EMAIL')
    responsible_id = BxField('RESPONSIBLE_ID')
    responsible_last_name = BxField('RESPONSIBLE_LAST_NAME')
    responsible_login = BxField('RESPONSIBLE_LOGIN')
    responsible_name = BxField('RESPONSIBLE_NAME')
    responsible_personal_photo = BxField('RESPONSIBLE_PERSONAL_PHOTO')
    responsible_second_name = BxField('RESPONSIBLE_SECOND_NAME')
    responsible_work_position = BxField('RESPONSIBLE_WORK_POSITION')
    status_id = BxField('STATUS_ID')
    tax_value = BxField('TAX_VALUE')
    company_id = BxField('UF_COMPANY_ID')
    contact_id = BxField('UF_CONTACT_ID')
    mycompany_id = BxField('UF_MYCOMPANY_ID')
    deal_id = BxField('UF_DEAL_ID')
    quote_id = BxField('UF_QUOTE_ID')
    user_description = BxField('USER_DESCRIPTION')
    invoice_properties = BxField('INVOICE_PROPERTIES')
    product_rows = BxField('PRODUCT_ROWS')

    _bx_meta = {
        'entity': 'crm.invoice',
        'default_prefix': 'FIELDS'
    }


class BxPaySystem(BxEntity):
    name = BxField('NAME')
    active = BxBoolean('ACTIVE')
    sort = BxField('SORT')
    description = BxField('DESCRIPTION')
    person_type_id = BxField('PERSON_TYPE_ID')
    action_file = BxField('ACTION_FILE')
    handler = BxField('HANDLER')
    handler_code = BxField('HANDLER_CODE')
    handler_name = BxField('HANDLER_NAME')

    _bx_meta = {
        'entity': 'crm.paysystem',
        'default_prefix': 'FIELDS'
    }


class BxPersonType(BxEntity):
    name = BxField('NAME')

    _bx_meta = {
        'entity': 'crm.persontype',
        'default_prefix': 'FIELDS'
    }


class BxProduct(BxEntity):
    name = BxField('NAME')
    active = BxBoolean('ACTIVE')
    preview_picture = BxField('PREVIEW_PICTURE')
    detail_picture = BxField('DETAIL_PICTURE')
    sort = BxField('SORT')
    xml_id = BxField('XML_ID')
    catalog_id = BxField('CATALOG_ID')
    section_id = BxField('SECTION_ID')
    description = BxField('DESCRIPTION')
    description_type = BxField('DESCRIPTION_TYPE')
    price = BxField('PRICE')
    currency_id = BxField('CURRENCY_ID')
    vat_id = BxField('VAT_ID')
    vat_included = BxField('VAT_INCLUDED')
    measure = BxField('MEASURE')

    _bx_meta = {
        'entity': 'crm.product',
        'default_prefix': 'FIELDS'
    }


class BxProductProperty(BxEntity):
    iblock_id = BxField('IBLOCK_ID')
    name = BxField('NAME')
    active = BxBoolean('ACTIVE')
    sort = BxField('SORT')
    default_value = BxField('DEFAULT_VALUE')
    property_type = BxField('PROPERTY_TYPE')
    row_count = BxField('ROW_COUNT')
    col_count = BxField('COL_COUNT')
    multiple = BxBoolean('MULTIPLE')
    xml_id = BxField('XML_ID')
    file_type = BxField('FILE_TYPE')
    link_iblock_id = BxField('LINK_IBLOCK_ID')
    is_required = BxBoolean('IS_REQUIRED')
    user_type = BxField('USER_TYPE')
    user_type_settings = BxField('USER_TYPE_SETTINGS')
    values = BxField('VALUES')

    _bx_meta = {
        'entity': 'crm.product.property',
        'default_prefix': 'FIELDS'
    }


class BxProductPropertyTypes(BxEntity):
    property_type = BxField('PROPERTY_TYPE')
    user_type = BxField('USER_TYPE')
    description = BxField('DESCRIPTION')

    _bx_meta = {
        'entity': 'crm.product.property.types'
    }
